/*

  This file is part of sniffer, a packet capture utility and
  network moniter
  The author can be contacted at <mistral@stev.org>
  the lastest version is avilable from 
  http://stev.org

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

	$Id: if.c,v 1.1 2009/05/06 17:07:00 james.stevenson Exp $


*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <pthread.h>
#include <time.h>
#include <fcntl.h>

#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/socket.h>

#include "config.h"
#include "locks.h"
#include "list.h"
#include "arp.h"
#include "sniff.h"
#include "stat.h"
#include "log.h"
#include "ip.h"
#include "if.h"

struct list_t *if_list;

/* The new way to read from the socket */
/* this should be able to tell that interface the packet came from */
/* and then send it to the correct handler */
int inline if_handle(struct sniff_pkt *pkt) {
	struct if_stat cmp; /* used for the searches */
	int i;


	list_LOCK(if_list);

	cmp.if_index = pkt->fromaddr.sll_ifindex;
	i = list_search(if_list, &cmp);
	if (i < 0) {
		pkt->face = if_add(pkt);
		if (!pkt->face)
			return -1; /* something bad happened */
	} else {
		pkt->face = list_get(if_list, i);
		if (!pkt->face)
			return -1; /* something bad happened */
	}

	list_UNLOCK(if_list); /* unlock the whole list */

	/* work out which packet type it is */
	switch(pkt->fromaddr.sll_hatype) {
		case LINK_ETHERNET: /* ethernet packet */
			return if_eth(pkt);
			break;
		case LINK_PPP:
			return if_raw(pkt);
			break;
		case ARPHRD_LOOPBACK:
			return if_lo(pkt);
			break;
		case LINK_SLIP:
		case 256:
		case ARPHRD_PPP:
			return if_raw(pkt);
			break;
		default:
			stat_global.dropped++;
			pkt->face->all_stat.dropped++;
			log_error("if_handle: Packet Type %d Not Known\n", pkt->fromaddr.sll_hatype);
			break;
	} /* end of hardware types */
	return -2;
}

/* This is called with the if_list LOCKED */
struct if_stat *if_add(struct sniff_pkt *pkt) {
	struct ifreq ifr;
	struct if_stat *tmp;
	int i, total;

	total = list_length(if_list);

	for(i=0;i<total;i++) {
		tmp = list_get(if_list, i);
		bzero(&ifr, sizeof(ifr));
		ifr.ifr_ifru.ifru_ivalue = pkt->fromaddr.sll_ifindex;
		if (ioctl(sockfd, SIOCGIFNAME, &ifr) < 0) {
			log_errno("ioctl");
		} else { /* do a string compare for if name */
			if (strcmp(&ifr.ifr_ifrn.ifrn_name[0], &tmp->if_name[0]) == 0) {
				list_del(if_list, i);
				tmp->if_index = pkt->fromaddr.sll_ifindex;
				list_add_sort(if_list, tmp);
				return tmp;
			}
		}
	}

	/* if we get down here create a new record */

	tmp = malloc( sizeof(struct if_stat));
	if (!tmp) {
		log_errno("malloc");
		return NULL;
	}

	init_stat(&tmp->all_stat, "ALL");
	init_stat(&tmp->in_stat,  "IN");
	init_stat(&tmp->out_stat, "OUT");
	init_stat(&tmp->other_stat, "OTHER");
	tmp->if_index = pkt->fromaddr.sll_ifindex;
	tmp->name = &tmp->if_name[0];

	bzero(&ifr, sizeof(ifr));
	ifr.ifr_ifru.ifru_ivalue = pkt->fromaddr.sll_ifindex;

	if (ioctl(sockfd, SIOCGIFNAME, &ifr) < 0) {
		log_errno("ioctl");
		sprintf(&tmp->if_name[0], "IF%d\n", pkt->fromaddr.sll_ifindex);
	}
	strcpy(&tmp->if_name[0], &ifr.ifr_ifrn.ifrn_name[0]);
	list_add_sort(if_list, tmp);
	return tmp;
}

int if_cmp_index(struct if_stat *c1, struct if_stat *c2) {
	if (c1->if_index > c2->if_index) return -1;
	if (c1->if_index < c2->if_index) return 1;
	if (c1->if_index == c2->if_index) return 0;
	return 0;
}

int if_init() {
	if_list = list_init();
	list_setcmp(if_list, (void *) if_cmp_index);
	return 0;
}

/* Functions for different interfaces */
int inline if_eth(struct sniff_pkt *pkt) {
	struct ethhdr *eth_packet; /* ethernet packet */

	if (pkt->dataleft < sizeof(struct ethhdr))
		return ERR_DATA;

	/* move forward */
	eth_packet = (struct ethhdr *) pkt->cur;
	pkt->cur += sizeof(struct ethhdr);
	pkt->dataleft -= sizeof(struct ethhdr);

	pkt->face->all_stat.packets++;
	pkt->face->all_stat.bytes += pkt->readlen;

	/* time to find out what ethernet packet it is */

	switch(pkt->fromaddr.sll_pkttype) {
		case PACKET_HOST:
			pkt->face->in_stat.packets++;
			pkt->face->in_stat.bytes += pkt->readlen;
			break;
		case PACKET_OUTGOING:
			pkt->face->out_stat.packets++;
			pkt->face->out_stat.bytes += pkt->readlen;
			break;
		default:
			pkt->face->other_stat.packets++;
			pkt->face->other_stat.bytes += pkt->readlen;
			break;
	}

	switch( ntohs(eth_packet->h_proto)) {  
		case ETH_P_ARP: /* arp packets */
			return arp_handle(pkt);
			break;
		case ETH_P_IP: /* ip - tcp/udp/icmp */
			return ip_handle(pkt);
			break;
		default:
			stat_global.dropped++;
			log_error("sniff_eth Unsupported Packet Type: %x\n", ntohs(eth_packet->h_proto));
			break;
	} /* end switch ethetnet packet type */
	return ERR_MISC;
}                       


int inline if_lo(struct sniff_pkt *pkt) {
	struct ethhdr *eth_packet; /* ethernet packet */

	if (pkt->dataleft < sizeof(struct ethhdr))
		return ERR_DATA;

	eth_packet = (struct ethhdr *) pkt->cur;
	pkt->cur += sizeof(struct ethhdr);
	pkt->dataleft -= sizeof(struct ethhdr);

	/* NOTE on a loop back interface we will see all
	data twice we dont want this so jump out ASAP */

	switch(pkt->fromaddr.sll_pkttype) {
		case PACKET_HOST:
			pkt->face->all_stat.packets++;
			pkt->face->all_stat.bytes += pkt->readlen;
			break;
		default:
			return 0;
			break;
	}

	/* time to find out what ethernet packet it is */
	switch( ntohs(eth_packet->h_proto)) {  
		case ETH_P_ARP: /* arp packets */
			return arp_handle(pkt);
			break;
		case ETH_P_IP: /* ip - tcp/udp/icmp */
			return ip_handle(pkt);
			break;
		default:
			stat_global.dropped++;
			log_error("sniff_eth Unsupported Packet Type: %x\n", ntohs(eth_packet->h_proto));
			break;
	} /* end switch ethetnet packet type */
	return ERR_MISC;
}                       


int inline if_raw(struct sniff_pkt *pkt) {

	switch(pkt->fromaddr.sll_pkttype) {
		case PACKET_HOST:
			pkt->face->in_stat.packets++;
			pkt->face->in_stat.bytes += pkt->readlen;
			break;
		case PACKET_OUTGOING:
			pkt->face->out_stat.packets++;
			pkt->face->out_stat.bytes += pkt->readlen;
			break;
		default:
			pkt->face->other_stat.packets++;
			pkt->face->other_stat.bytes += pkt->readlen;
			break;
	}

	pkt->face->all_stat.packets++;
	pkt->face->all_stat.bytes += pkt->readlen;

	switch(ntohs(pkt->fromaddr.sll_protocol)) {
		case ETH_P_IP:
			return ip_handle(pkt);
			break;
		default:
			log_error("sniff_raw Unsupported Packet: %x\n", ntohs(pkt->fromaddr.sll_protocol));
			return ERR_SUPP;
			break;
	} /* end of different protocls */
	return ERR_MISC;
}

