/*

  This file is part of sniffer, a packet capture utility and
  network moniter
  The author can be contacted at <mistral@stev.org>
  the lastest version is avilable from 
  http://stev.org

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

	$Id: ip.c,v 1.1 2009/05/06 17:07:02 james.stevenson Exp $

*/

#include "config.h"

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <time.h>
#include <pthread.h>

#include <netinet/in.h>

#include "log.h"
#include "stat.h"
#include "in_ntoa.h"
#include "lookup.h"
#include "ip.h"
#include "tcp.h"
#include "udp.h"
#include "icmp.h"

struct gen_stat ip_stat;
int (*ip_functions[255]) (struct sniff_pkt *pkt, struct pkt_ip *ip);
int ip_support[255];
FILE *ip_log = NULL;

int ip_register(unsigned char proto, void *func) {
	if (ip_functions[proto] != NULL)
		return -1;
	ip_functions[proto] = func;
	return 0;
}

int ip_unregister(unsigned char proto) {
	ip_functions[proto] = NULL;
	return 0;
}

int ip_handle(struct sniff_pkt *pkt) {
	struct pkt_ip *ip;
	unsigned short len;
	char *ip_source = NULL, *ip_dest = NULL;

#ifdef IP_CHECKSUM
	unsigned short oldcheck; /* used to store the ckeck sum when we check it */
#endif

	ip_stat.packets++;
	ip_stat.bytes += pkt->dataleft;

	if (pkt->dataleft < sizeof(struct pkt_ip)) {
		ip_stat.dropped++;
		return ERR_DATA;
	}
	ip = pkt->cur;

	/* sanity checks */
	if (ip->version != 4) {
		fprintf(ip_log, "Bad IP Version\n");
		ip_stat.dropped++;
		return ERR_MISC;
	}

	if (ip->ihl < 5) {
		fprintf(ip_log, "IP header length to short for valid packet\n");
		ip_stat.dropped++;
		return ERR_MISC;		
	}

	len = ip->ihl * 4;
	pkt->dataleft -= len;
	if (pkt->dataleft < 0) {
		ip_stat.dropped++;
		return ERR_DATA;
	}
	pkt->cur += len;

#ifdef IP_CHECKSUM
	/* do the checksum calcuation */
	oldcheck = ip->check;
	ip->check = 0;
	ip->check = ip_cksum( (unsigned short *) ip, len);

	if ( oldcheck != ip->check) {
		ip_source = lookup(ip->saddr, 0);
		ip_dest = lookup(ip->daddr, 0);
		fprintf(ip_log, "FAILED IP CHECKSUM: ?%s -> ?%s\n", ip_source, ip_dest);
		ip_stat.dropped++;
		return ERR_CSUM;
	}
#endif

	lookup_stats(ip->daddr, pkt->dataleft, ip->saddr, pkt->dataleft);

	if ( ntohs(ip->frag_off) & (IP_MF|IP_OFFSET) ) {
		ip_stat.dropped++;
		return ERR_SUPP;
	}

	if (ip_functions[ip->protocol] != NULL)
		return ip_functions[ip->protocol](pkt, ip);

	if (!ip_support[ip->protocol]) {
		ip_source = lookup(ip->saddr, 0);
		ip_dest = lookup(ip->daddr, 0);
		fprintf(ip_log, "%s -> %s IPPROTO: %u\n", ip_source, ip_dest, ip->protocol);
		ip_support[ip->protocol] = 1;
	}

	return ERR_SUPP;
} /* end of ip_handle */


inline unsigned short ip_cksum(unsigned short *buff, int len) {
	unsigned long sum = 0;
	while(len > 1) {
		sum += *buff++;
		len -= 2;
	}
	if (len == 1)
		sum += (*buff & 0xff);
	sum = (sum >> 16) + (sum & 0xffff);
	sum += (sum >> 16);
	sum = ~sum;
	return (sum & 0xffff);
}

int ip_init() {
	int i;
	for(i=0;i<255;i++)
		ip_functions[i] = NULL;
	for(i=0;i<255;i++)
		ip_support[i] = 0;
	ip_log = fopen("output/ip_log", "w+");
	if (!ip_log) {
		log_error("Unable to open ip log file\n");
		return -1;
	}
	return 0;
}

